# -*- coding: utf-8 -*-
"""

 MODULE:       hru-delin_basins

 AUTHOR(S):    adapted from GRASS-HRU (ILMS) - JENA University
               by IRSTEA - Christine Barachet 

 PURPOSE:    1. Relocates the gauges on the reaches
             2. Calculates watersheds at the gauges

"""

import subprocess, string, os, sys, glob, types, time
import numpy as np
import ConfigParser
import grass.script as grass
import struct
import math
import csv

from osgeo import gdal
from gdalconst import *
from osgeo import ogr


def write_log(msg, buff, start, end, best_x, best_y):
    horiz_shift = best_x - start - end 
    vertic_shift = best_y - start - end + 1
    if horiz_shift != 0 or vertic_shift != 0:
        if horiz_shift > 0:
            dir_x_shift = 'East'
        else:
            dir_x_shift = 'West'
    
        if vertic_shift > 0:
            dir_y_shift = 'South'
        else:
            dir_y_shift = 'North'
    
        logf.writerow({'gauge code' : gauge_code, 'area' : str(gauge_area), 'rule' :  msg, 
                       'horiz dir' : dir_x_shift,
                       'horiz shift' : str(abs(horiz_shift)),
                       'vertic dir' : dir_y_shift, 
                       'vertic shift' : str(abs(vertic_shift))})
                       

def snapping_gauges_to_reaches():
    """
    Relocates gauges on the reaches derivated from MNT in step 1 (r.watershed)
    """
    basin_min_size = int(parms.get('basin_min_size', 'size'))
    surface_ok_1 = int(parms.get('auto_relocation', 'surface_tolerance_1'))
    dist_ok_1 = int(parms.get('auto_relocation', 'distance_tolerance_1' ))

    if parms.get('auto_relocation', 'surface_tolerance_2') == "":
        nb_trials = 1
        surface_ok_2 = 0
        dist_ok_2 = 0
    else:
        nb_trials = 2
        surface_ok_2 = int(parms.get('auto_relocation', 'surface_tolerance_2'))
        dist_ok_2 = int(parms.get('auto_relocation', 'distance_tolerance_2' ))
    
    col_area = parms.get('auto_relocation', 'gauge_area_col_name')
    if col_area == "":
        print(' =========> ERROR ')
        sys.exit(' =========> CHECK gauge_area_col_name PARAMETER')

    gauge_area_unit = parms.get('auto_relocation', 'gauge_area_unit')
    # are these the right factors?
    # if values is entered in square meters, fact should be 1, no?
    if gauge_area_unit == '1':
        fact = 1000
        unit = 'm2'
    elif gauge_area_unit == '2':
        fact = 1000000
        unit = 'km2'
    else:
        print(' =========> ERROR ')
        sys.exit(' =========> CHECK gauge_area_unit PARAMETER')

    global logf, gauge_code, gauge_area
    logfile_n = directory_out + '/auto_reloc.csv'
    logfile = open(logfile_n, 'w')
    fieldnames = ['gauge code', 'area', 'rule', 'horiz shift', 'horiz dir', 'vertic shift', 'vertic dir']
    logf = csv.DictWriter(logfile, fieldnames=fieldnames, delimiter=';')
    logf.writeheader()


    # Read the accumulation raster and get the geo informations
    accum_file = directory_out + '/step1_accum.tif'
    accum_lyr = gdal.Open(accum_file, GA_ReadOnly)
    accum_bd = accum_lyr.GetRasterBand(1)
    georef = accum_lyr.GetGeoTransform()
    hres = georef[1]
    vres = georef[5]
    pixel_sz = (hres * abs(vres))
    # It gets the number of rows and columns of the accumulation raster file
    acc_rast_cols = accum_lyr.RasterXSize
    acc_rast_rows = accum_lyr.RasterYSize
       
    # Matrix of euclidian distances from the center (gauge position)
    max_dist = max(dist_ok_1, dist_ok_2)
    euclid_dist = np.full((max_dist * 2 + 1, max_dist * 2 + 1), 0, dtype=int)
    for y in range(0, (max_dist * 2 + 1)):
        for x in range(0, (max_dist * 2 + 1)):
            euclid_dist[y, x] = int(math.sqrt((y - max_dist)**2 + (x - max_dist)**2))
  
    # Gauges loop                                    
    for gauge in reloc_lyr:
        gauge_code = gauge.GetField(col_name)
        # Get the drained surface of the gauge
        gauge_area = gauge.GetField(col_area)

        # ------------------------------------------------------------
        # if gauge_area < 0:
        #   nb_pixel_area = basin_min_size + 1
        #   ignore_surface_tolerance = True
        # else:
        #   nb_pixel_area = ((gauge_area * fact) / pixel_sz) + 0.5
        #   ignore_surface_tolerance = False
        # ------------------------------------------------------------
        # here, the corresponding number of cell of the specified drained area is calculated
        nb_pixel_area = ((gauge_area * fact) / pixel_sz) + 0.5
        # ------------------------------------------------------------

        if nb_pixel_area > basin_min_size:
            # Get the coordinates of the gauge and calculate center of the offset
            # I think it relocates the station in the center of the closest pixel
            # in the raster
            geom = gauge.GetGeometryRef()
            gauge_x, gauge_y = geom.GetX(), geom.GetY()
            Xreste = gauge_x % hres
            Yreste = gauge_y % abs(vres)
            new_gauge_x = (gauge_x - Xreste) + 1
            new_gauge_y = (gauge_y - Yreste) - 1
          
            # Read the currents of accum raster around the gauge
            # (new_gauge_x - georef[0]) / hres)
            # ==> this gives the column in raster file where the station has been relocated
            # int(((new_gauge_x - georef[0]) / hres) - (dist_max))
            # ==> get the difference with the maximum number of column of accepted differences
            #     if > 0 then, dist_max is < column number of gaug
            #     if < 0 then, dist_max is > column number of gaug
            # max(..., 0) ==> there, we just see if gaug is or not within a dist_max distance from left border???

            dist_max = max(dist_ok_1, dist_ok_2)
            pix_x = max(int(((new_gauge_x - georef[0]) / hres) - (dist_max)), 0)
            pix_y = max(int(((georef[3] - new_gauge_y) / abs(vres)) - (dist_max)), 0)

            #        and don't access beyond the accum raster bounds
            nb_cols = min((dist_max * 2) + 1, acc_rast_cols - pix_x)
            nb_rows = min((dist_max * 2) + 1, acc_rast_rows - pix_y)
            zone = accum_bd.ReadAsArray(pix_x, pix_y, nb_cols, nb_rows).astype(int)
            # in the end, this is just to ensure that we won't look at relocating the gaug outside the raster boudaries
            # isn't that right?
            # zone contains the accumulation raster data within max_dist area

            # Matrix of differences
            zone_diff = abs(nb_pixel_area - abs(zone))
            # here, the differences between the actual drained area and the specified drained area is calculated

            # Prepare for first search
            # here accum_toler is the number of difference in pixel that can be tolerated
            accum_toler = int(nb_pixel_area * surface_ok_1 / 100)
            if dist_ok_1 > dist_ok_2:
                start = 0
                col_max = min(nb_cols, acc_rast_cols - pix_x)
                row_max = min(nb_rows, acc_rast_rows - pix_y)
                end = dist_ok_1 - start
            else:
                start = dist_ok_2 - dist_ok_1
                col_max = min(nb_cols - start, acc_rast_cols - pix_x)
                row_max = min(nb_rows - start, acc_rast_rows - pix_y)
                end = dist_ok_2 - start
            msg = 'shifted with first rule'
            buff = dist_ok_1
            trial = 1

            while trial <= nb_trials:
                # best_area is useless... not used anywhere...
                best_area = zone_diff.max()
                best_euclid_dist = euclid_dist[start, start]
                best_x = 0
                best_y = 0
                better_place_found = False
               
                for y in range(start, row_max):
                    for x in range(start, col_max):
                        # ------------------------------------------------------------
                        # if ignore_surface_tolerance:
                        #   if euclid_dist[y, x] < best_euclid_dist:
                        #        best_euclid_dist = euclid_dist[y, x]
                        #        best_y = y
                        #        best_x = x
                        #        better_place_found = True
                        # else:
                        #    if zone_diff[y, x] <= accum_toler:
                        #        if euclid_dist[y, x] < best_euclid_dist:
                        #            best_euclid_dist = euclid_dist[y, x]
                        #            best_y = y
                        #            best_x = x
                        #            better_place_found = True
                        # ------------------------------------------------------------
                        if zone_diff[y, x] <= accum_toler:
                            if euclid_dist[y, x] < best_euclid_dist:
                                # best_area is useless... not used anywhere...
                                best_area = zone_diff[y, x]
                                best_euclid_dist = euclid_dist[y, x]
                                best_y = y
                                best_x = x
                                better_place_found = True
                        # ------------------------------------------------------------
                if better_place_found:
                    new_x = ((pix_x + best_x) * hres) + georef[0] + (hres / 2)
                    new_y = georef[3] - ((pix_y + best_y) * abs(vres)) - (abs(vres) / 2)
                    wkt   = "POINT(%f %f)" %  (new_x , new_y)
                    point = ogr.CreateGeometryFromWkt(wkt)
                    gauge.SetGeometryDirectly(point)
                    reloc_lyr.SetFeature(gauge)
                    trial = 3
                    write_log(msg, buff, start, end, best_x, best_y)
                else:
                    if trial == 2:
                        logf.writerow({'gauge code' : gauge_code, 'area' : str(gauge_area), 'rule' : ' no better place found'})
                        trial = 3
                    else:
                        if nb_trials == 1:
                            logf.writerow({'gauge code' : gauge_code, 'area' : str(gauge_area), 'rule' : ' no better place found'})
                            trial = 3
                        else:
                            # Prepare for second search
                            msg = 'shifted with second rule'                            
                            accum_toler = int(nb_pixel_area * surface_ok_2 / 100) 
                            trial = 2
                            buff = dist_ok_2
                            if dist_ok_2 > dist_ok_1:
                                start = 0
                                end = dist_ok_2 - start
                            else:
                                start = dist_ok_1 - dist_ok_2
                                col_max = min(nb_cols - start, acc_rast_cols - pix_x)
                                row_max = min(nb_rows - start, acc_rast_rows - pix_y )
                                end = dist_ok_1 - start
        else:
           gaugeID = gauge.GetFID()
           reloc_lyr.DeleteFeature(gaugeID)
           logf.writerow({'gauge code' : gauge_code, 'area' : str(gauge_area), 'rule' : ' basin too small ---> DELETED'})
   
    logfile.close

def cut_streams_at_gauges():
    """
    Streams generated at step 1 (by r.watershed) are cutted at the gauges
    """
    streams_file = directory_out + '/step1_streams.tif'
    streams_lyr = gdal.Open(streams_file, GA_ReadOnly)
    streams_band = streams_lyr.GetRasterBand(1)
    stream_orig = streams_band.ReadAsArray()
    
    mask_file = directory_out + '/step2_mask.tif'
    mask_lyr = gdal.Open(mask_file, GA_ReadOnly)
    mask_band = mask_lyr.GetRasterBand(1)
    mask = mask_band.ReadAsArray()
    nodata = mask_band.GetNoDataValue()
    mask[mask == nodata] = 0

    stream_orig = stream_orig * mask
 
    streams_cols = streams_lyr.RasterXSize
    streams_rows = streams_lyr.RasterYSize
    georef = streams_lyr.GetGeoTransform()
    hres = georef[1]
    vres = georef[5]

    streams_new_file = directory_out + '/step2_streams_new.tif'
    driver = gdal.GetDriverByName('GTiff')
    dst_ds = driver.Create(streams_new_file, streams_cols, streams_rows, 1, GDT_Int32)
    dst_ds.SetGeoTransform((georef[0], hres, 0, georef[3], 0, vres))
    dst_ds.SetProjection(streams_lyr.GetProjection())

    drain_file = directory_out + '/step1_drain.tif'
    drain_lyr = gdal.Open(drain_file, GA_ReadOnly)
    drain = drain_lyr.GetRasterBand(1).ReadAsArray()
    drain_dir = [[-1,1],[-1,0],[-1,-1],[0,-1],[1,-1],[1,0],[1,1],[0,1]]

    # Give possibility of more than 2 segments in a stream
    stream_wk = np.zeros((streams_rows, streams_cols), dtype=('int64'))
    nodata = streams_band.GetNoDataValue()
    for y in range (0, streams_rows - 1):
        for x in range (0, streams_cols - 1):    
            if stream_orig[y, x] != nodata:
                stream_wk[y, x] = stream_orig[y, x] * 100

    gauges_file = directory_out + '/step2_gauges_for_watersheds.shp'
    gauges_in = ogr.Open(gauges_file, 1)
    gauges_lyr = gauges_in.GetLayer()

    for gauge in gauges_lyr:
        gauge_code = gauge.GetField(col_name)
        geom = gauge.GetGeometryRef()
        current_x = int(((geom.GetX() - georef[0]) / hres))
        current_y = int(((georef[3] - geom.GetY()) / abs(vres)))
        current_stream = stream_orig[current_y, current_x]
        
        if current_stream == nodata:
            print('---------- gauge outside of stream network : ', gauge_code)
            end_of_stream = True
        else:
            stream_wk[current_y, current_x] += 1
            end_of_stream = False
            i = 0
            d = abs(drain[current_y, current_x])

        while not end_of_stream:
            if d == 8:
                d = 0
            y = current_y + drain_dir[d][0]
            x = current_x + drain_dir[d][1]
            d2 = abs(drain[y, x]) - 1
            drain_into_y = y + drain_dir[d2][0]
            drain_into_x = x + drain_dir[d2][1]
            if (stream_orig[y, x] == current_stream) and (drain_into_y == current_y) and (drain_into_x == current_x):
                stream_wk[y, x] += 1
                current_y = y
                current_x = x
                i = 0
                d = abs(drain[y, x])
            else:
                i += 1
                if i == 7:
                    end_of_stream = True    
                else:
                    d += 1
    # ----------------------------------------------------------------------------------
    # the following code was added by I.Horner (2018-03-28), later modified (2018-06-19)
    gauges_file = directory_out + '/step2_gauges_for_watersheds.shp'
    gauges_in = ogr.Open(gauges_file, 1)
    gauges_lyr = gauges_in.GetLayer()
    gauges_reach_id = [] # added by I.Horner (2018-03-28) to get reach id for each gauging station
    for gauge in gauges_lyr:
        gauge_code = gauge.GetField(col_name)
        # print("--------------------------------------------------------")
        # print("Processing gauging station "+str(gauge_code))
        geom = gauge.GetGeometryRef()
        current_x = int(((geom.GetX() - georef[0]) / hres))
        current_y = int(((georef[3] - geom.GetY()) / abs(vres)))
        # print("x = "+str(current_x))
        # print("y = "+str(current_y))
        current_stream = stream_wk[current_y, current_x]
        # print("Corresponding reach id: "+str(current_stream))
        gauges_reach_id.append(current_stream)
        
    # to write reach id corresponding to each gauging station directly in attribute table
    dst_ds.GetRasterBand(1).WriteArray(stream_wk)
    dst_ds = None
    gauges_file = directory_out + '/step2_gauges_for_watersheds.shp'
    gauges_in = ogr.Open(gauges_file, 1)
    gauges_lyr = gauges_in.GetLayer()
    gauges_lyr.CreateField(ogr.FieldDefn("ReachID", ogr.OFTInteger))
    k = 0
    gauges_features = gauges_lyr.GetNextFeature()
    while gauges_features:
        gauges_features.SetField("ReachID", int(gauges_reach_id[k]))
        gauges_lyr.SetFeature(gauges_features)
        gauges_features = gauges_lyr.GetNextFeature()
        k = k + 1
    # Set the new shape
    dir_station_out = directory_out + '/stations.shp' #
    # gauges_reloc_name = 'step2_gauges_for_watersheds'
    # gauges_reloc_file = directory_out + '/' + gauges_reloc_name + '.shp'
    driver = ogr.GetDriverByName('ESRI Shapefile') # returns a Driver object
    if os.path.exists(dir_station_out):
        driver.DeleteDataSource(dir_station_out)
    station_shp = driver.CreateDataSource(dir_station_out) # Returns a DataSource object
    station_lyr = station_shp.CopyLayer(gauges_lyr, gauges_reloc_name) 

'''

 MAIN


'''
try:
    from subprocess import DEVNULL # Python 3.
except ImportError:
    DEVNULL = open(os.devnull, 'wb')
    #DEVNULL = open('trace_step2', 'wb')
print('---------- HRU-delin Step 2 started ---------------------------------------------')

global logf, gauge_code, gauge_area, reloc_lyr, col_name

parms = ConfigParser.ConfigParser(allow_no_value=True)
parms_file = "hrudelin_config.cfg"
parms.read(parms_file)
directory_out = parms.get("dir_out", "files")

col_name = parms.get('for_watershed_id', 'col_name')
if col_name == "":
    print(' =========> ERROR ')
    sys.exit(' =========> CHECK col_name PARAMETER')

# Get the shape of gauges 
gauges_file = parms.get("auto_relocation", "relocated_gauges")
if gauges_file == "":
    gauges_file = directory_out + '/gauges_selected.shp'
gauges_in = ogr.Open(gauges_file)
gauges_lyr = gauges_in.GetLayer()
 
# Set the new shape
gauges_reloc_name = 'step2_gauges_for_watersheds'
gauges_reloc_file = directory_out + '/' + gauges_reloc_name + '.shp'
driver = ogr.GetDriverByName('ESRI Shapefile') 
if os.path.exists(gauges_reloc_file):
    driver.DeleteDataSource(gauges_reloc_file)
reloc_shp = driver.CreateDataSource(gauges_reloc_file)
reloc_lyr = reloc_shp.CopyLayer(gauges_lyr, gauges_reloc_name)

# Relocation of the gauges
if (parms.get("auto_relocation", "to_do")) == 'yes':
    print('---------- HRU-delin Step 2 : Relocation of the gauges')
    snapping_gauges_to_reaches()
reloc_shp.ExecuteSQL('REPACK ' + reloc_lyr.GetName())
reloc_shp.Destroy()

# Import drain raster
drain_layer = directory_out + '/step1_drain.tif'
drain_wk = 'drain_wk'
grass.run_command('r.in.gdal', flags='o', input=drain_layer, output=drain_wk, overwrite='True', stdout = DEVNULL, stderr = DEVNULL)
grass.run_command('g.proj', flags='p', georef='drain_wk', stdout = DEVNULL, stderr = DEVNULL)
grass.run_command('g.region', flags='sp', rast=drain_wk, stdout = DEVNULL, stderr = DEVNULL)

# Watersheds derivation 
basins='basins'
rasters_list='basins'
# r.cross doesn't accept more than 30 layers
max_rasters=29
i=0
grass.run_command('r.mapcalc', basins='null()', stdout = DEVNULL, stderr = DEVNULL)
ds = ogr.Open(gauges_reloc_file)
reloc_lyr = ds.GetLayer()
nb_gauges = reloc_lyr.GetFeatureCount()

print('---------- HRU-delin Step 2 : Derivating watersheds at gauges')
for gauge in reloc_lyr:
    geom = gauge.GetGeometryRef()
    gauge_x, gauge_y = geom.GetX(), geom.GetY()
    basin_wk = 'basin_wk'
    grass.run_command('r.water.outlet', drainage=drain_wk, basin='basin_tmp%d' %(i),
                       easting = gauge_x, northing = gauge_y, overwrite='True', stdout = DEVNULL, stderr = DEVNULL)
    rasters_list += ',basin_tmp%d' %(i)
    i += 1
    if i % max_rasters == 0 or i == nb_gauges:
        grass.run_command('r.cross', input=rasters_list, output=basins, overwrite='True', stdout = DEVNULL, stderr = DEVNULL)
        grass.run_command('g.remove', flags='f', rast='basin_tmp*', stdout = DEVNULL, stderr = DEVNULL)
        rasters_list='basins'  
  
grass.run_command('r.null', map=basins, setnull=0, stdout = DEVNULL, stderr = DEVNULL)

print('---------- HRU-delin Step 2 : Reclassifying the watersheds')
recl_watersheds_file = 'reclass_watersheds_rules'
rules_ws = open(recl_watersheds_file, "w")
ds = ogr.Open(gauges_reloc_file)
reloc_lyr = ds.GetLayer()

for gauge in reloc_lyr:
    geom = gauge.GetGeometryRef()
    gauge_x, gauge_y = geom.GetX(), geom.GetY()
    out = grass.read_command('r.what', input=basins, east_north=(gauge_x, gauge_y), stdout = DEVNULL, stderr = DEVNULL)
    cat_grass = out.split('|')[3].strip('\n')
    rules_ws.write(cat_grass.strip() + ':' + cat_grass.strip() + ':' + str(int(gauge.GetField(col_name))) + '\n')
rules_ws.close()
basins_rcl='basins_rcl'
cmd = 'cat '+recl_watersheds_file+' | r.recode --o input='+basins+' output='+basins_rcl
p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT) 
p.communicate()
# Save the rules to directory_out
cmd = 'mv '+recl_watersheds_file+' '+directory_out
subprocess.call(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
grass.run_command('r.out.gdal', input=basins_rcl, type='UInt16', output=directory_out + '/' + 'step2_watersheds.tif', overwrite='True', stdout = DEVNULL, stderr = DEVNULL)

# Create mask = surface of all watersheds
print('---------- HRU-delin Step 2 : Creating the mask')
grass.run_command('r.null', map=basins_rcl, setnull=0, stdout = DEVNULL, stderr = DEVNULL)
mask_tmp = 'mask_tmp'
cmd = 'echo "0 thru 10000000 = 1" | r.reclass --o input='+basins_rcl+' output='+mask_tmp
p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
p.communicate()
grass.run_command('r.null', map=mask_tmp, null=0, setnull=1, stdout = DEVNULL, stderr = DEVNULL)
grass.run_command('r.out.gdal', input=mask_tmp, type='UInt16', output=directory_out + '/' + 'step2_mask.tif', overwrite='True', stdout = DEVNULL, stderr = DEVNULL)

# Cutting the streams at gauges 
print('---------- HRU-delin Step 2 : Cutting the streams at gauges')
cut_streams_at_gauges()

# This last bit of code was added by I. Horner.
# > A new subassin layer is computed by the cross over of subassin and watershed.
# > The resulting features are attributed the same ids as the reaches
# > Everything is saved as raster AND shapefile after running r.to.vect, for easy plotting.
print('---------- HRU-delin Step 2 : Computing subbasinXwatershed raster layer')

dem_recl = directory_out + '/' + 'step1_dem_reclass.tif'
grass.run_command('g.proj', flags='p', georef=dem_recl, stdout = DEVNULL, stderr = DEVNULL)

print('---------- Importing raster \'step1_subbasins.tif\'')
subbasins = directory_out + '/' + 'step1_subbasins.tif'
grass.run_command('r.in.gdal', flags='o', input=subbasins, output='subbasins', overwrite='True', stdout = DEVNULL, stderr = DEVNULL)

print('---------- Importing raster \'step2_watersheds.tif\'')
basins = directory_out + '/' + 'step2_watersheds.tif'
grass.run_command('r.in.gdal', flags='o', input=basins, output='watersheds', overwrite='True', stdout = DEVNULL, stderr = DEVNULL)

print('---------- Importing raster \'step2_streams_new.tif\'')
reachraster = directory_out + '/' + 'step2_streams_new.tif'
grass.run_command('r.in.gdal', flags='o', input=reachraster, output='reachraster', overwrite='True', stdout = DEVNULL, stderr = DEVNULL)

print('---------- Compute cross product \'watersheds*100000+subbasins\'')
grass.run_command('r.mapcalc', cross1='watersheds*100000+subbasins', stdout = DEVNULL, stderr = DEVNULL)

print('---------- Compute cross product cross1*reachraster')
grass.run_command('r.mapcalc', cross2='if(reachraster!=0, cross1, 0)', stdout = DEVNULL, stderr = DEVNULL)

print('---------- Setting nulls in \'cross1*reachraster\' and \'reachraster\'')
grass.run_command('r.null', map='cross2', setnull=0, stdout = DEVNULL, stderr = DEVNULL)
grass.run_command('r.null', map='reachraster', setnull=0, stdout = DEVNULL, stderr = DEVNULL)

print('---------- Saving \'step2_subbasins.tif\' (watersheds*100000+subbasins)')
grass.run_command('r.out.gdal', input='cross1', output=directory_out + '/' + 'step2_subbasins.tif', overwrite='True', stdout=DEVNULL, stderr=DEVNULL)

print('---------- Computing links between (watersheds*100000+subbasins) and reach ids')
reach_ids = os.popen("r.stats -nN --quiet reachraster").readlines()
reach_ids_cleaned = []
subbasins_cleaned = []
n_reach = len(reach_ids)
i = 1
for k in reach_ids:
    print("\r["+str(i)+"/"+str(n_reach)+"] Retrieving reach ids corresponding subbasins        "),
    sys.stdout.flush()
    os.system("r.mask -o --quiet input=reachraster maskcats=%s" % k.rstrip("\n"))
    tmp_val = os.popen("r.stats -nN --quiet cross1").readlines()
    reach_ids_cleaned.append(k.rstrip("\n"))
    subbasins_cleaned.append(tmp_val[0].rstrip("\n"))
    i+=1
print("")
os.system("r.mask -r --quiet MASK")
if os.path.isfile("tmp/reclass_reach.txt"):
        os.remove("tmp/reclass_reach.txt")
file = open("tmp/reclass_reach.txt", "w") 
for k in range(0, len(reach_ids_cleaned)):
    file.write(subbasins_cleaned[k]+"="+reach_ids_cleaned[k]+"\n")
file.close() 

print('---------- Reclassifying (watersheds*100000+subbasins)')
os.system("r.reclass --o input=cross1 output=cross1_reclassed rules='tmp/reclass_reach.txt'")

print('---------- Saving \'step2_subbasins.tif\'')
grass.run_command('r.out.gdal', input='cross1_reclassed', output=directory_out + '/' + 'step2_subbasins_2.tif', overwrite='True', stdout=DEVNULL, stderr=DEVNULL)

print('---------- Creating vector layers from raster layers ... ')
grass.run_command('r.to.vect', flags='v', quiet=True, input="subbasins", output='subbasins_vector', feature='area', overwrite='True')
grass.run_command('r.to.vect', flags='v', quiet=True, input="watersheds", output='watersheds_vector', feature='area', overwrite='True')
grass.run_command('r.to.vect', flags='v', quiet=True, input="cross1", output='cross1_vector', feature='area', overwrite='True')
grass.run_command('r.to.vect', flags='v', quiet=True, input="cross1_reclassed", output='cross1_reclassed_vector', feature='area', overwrite='True')

grass.run_command('v.out.ogr', flags='c', quiet=True, input='subbasins_vector', type='area', dsn=directory_out+'/step2_step1_subbasins.shp')
grass.run_command('v.out.ogr', flags='c', quiet=True, input='watersheds_vector', type='area', dsn=directory_out+'/step2_step2_watersheds.shp')
grass.run_command('v.out.ogr', flags='c', quiet=True, input='cross1_vector', type='area', dsn=directory_out+'/step2_step2_subbasins.shp')
grass.run_command('v.out.ogr', flags='c', quiet=True, input='cross1_reclassed_vector', type='area', dsn=directory_out+'/step2_step2_subbasins_2.shp')

print('---------- HRU-delin Step 2 ended    ---------------------------------------------')
